<?php

namespace BrewPixel\Barista\bin\composer;

use Composer\Script\Event;
use Composer\Command\ShowCommand;
use BrewPixel\Barista\Percolator\Validate;
use BrewPixel\Barista\Percolator\Sanitise;
use LayerShifter\TLDExtract\Extract;
use Dotenv\Dotenv;

/**
 * Class PostCreateProjectScript
 * @package BrewPixel\Roastery
 */
class PostCreateProjectScript {

	// @TODO: Test to see if these are ok as private.
	private static $client_name;
	private static $project_root;
	private static $choice;
	private static $main_repo_name;
	private static $domain_without_tld;
	private static $sship;
	private static $sshuser;
	private static $theme_name;
	private static $theme_name_lowercase_hyphen_no_spaces;
	private static $db_name;
	private static $db_user;
	private static $db_password;
	private static $db_host;
	private static $bitbucket_consumer_key;
	private static $bitbucket_consumer_secret;
	private static $bitbucket_team_name;
	private static $bitbucket_project_name;
	private static $bitbucket_project_key;
	private static $bitbucket_project_description;
	private static $site_url;
	private static $site_title_blog_name;
	private static $blog_description;
	private static $composer_project;
	private static $ignore_themes;
	private static $ignore_plugins;
	private static $ignore_wp;
	private static $all_fake_content;
	private static $fake_comments;
	private static $fake_posts;
	private static $fake_client_user;
	private static $development_mode;
	private static $simple;
	private static $admin_user;
	private static $maintenance;

	public function __construct() {
	}

	public static function bashInstall( Event $event ) {

		$dotenv = new Dotenv( dirname( dirname( dirname( dirname( dirname(__DIR__) ) ) ) ) );
		$dotenv->load();
		self::$maintenance = getenv('MAINTENANCE_PKG');

		$io                 = $event->getIO();
		self::$project_root = dirname( __DIR__ . DIRECTORY_SEPARATOR . '../../../../../..' );

		$project_new_wordpress_genesis                    = 'New Project: <info>WordPress Genesis-based</info>';
		$project_new_wordpress_genesis_woocommerce        = 'New Project: <info>WordPress Genesis-based WooCommerce</info>';
		$project_new_wordpress_moka                       = 'New Project: <info>WordPress Moka based</info>';
		$project_new_wordpress_moka_woocommerce           = 'New Project: <info>WordPress Moka based WooCommerce</info>';
		$install_reinstall_wordpress_current_project      = 'Existing Project: <info>Install/Re-install WordPress</info> <comment></comment>';
		$maintenance_update_core_plugins_themes_languages = 'Maintenance: <info>Update Core/Plugins/Themes/Languages</info> <comment></comment>';
		$advanced                                         = 'Advanced: <comment>Coming soon</comment>';
		$list_repos                                       = 'List all BrewPixel BitBucket Repos: <comment>Needs cleaned and paginated.</comment>';
		$sync_production_database_to_local                = 'Sync Production database to local: <comment>One direction production -> local.</comment>';
		$sync_production_database_to_staging              = 'Sync Production database to staging: <comment>One direction production -> staging.</comment>';

		self::$choice = $io->select(
			"\nWhat would you like to do?\n",
			array(
				0 => "$project_new_wordpress_genesis",
				1 => "$project_new_wordpress_genesis_woocommerce",
				2 => "$project_new_wordpress_moka",
				3 => "$project_new_wordpress_moka_woocommerce",
				4 => "$install_reinstall_wordpress_current_project",
				5 => "$maintenance_update_core_plugins_themes_languages",
				6 => "$advanced",
				7 => "$list_repos",
				8 => "$sync_production_database_to_local",
				9 => "$sync_production_database_to_staging",
			),
			'0',
			false,
			'Value "%s" is invalid',
			false
		);

		if ( $sync_production_database_to_local === self::$choice ) {
			self::sshQuestions( $io, self::$choice );
			putenv( 'SSH_IP=' . self::$sship );
			putenv( 'SSH_USER=' . self::$sshuser );
			$extract            = new Extract;
			$domain_extract     = $extract->parse( self::$main_repo_name );
			self::$domain_without_tld = preg_replace( '/\.' . $domain_extract['suffix'] . '$/',
				'', $domain_extract );
			putenv( 'DOMAIN_WITHOUT_TLD=' . self::$domain_without_tld );

			$production_db_to_local_db = __DIR__ . '/../wp/sync_production_db_to_local_db.sh';
			exec( "bash $production_db_to_local_db", $dboutput, $dbwpout );
		}

		if ( $sync_production_database_to_staging === self::$choice ) {
			self::sshQuestions( $io, self::$choice );
			putenv( 'SSH_IP=' . self::$sship );
			putenv( 'SSH_USER=' . self::$sshuser );
			$extract            = new Extract;
			$domain_extract     = $extract->parse( self::$main_repo_name );
			self::$domain_without_tld = preg_replace( '/\.' . $domain_extract['suffix'] . '$/',
				'', $domain_extract );
			putenv( 'DOMAIN_WITHOUT_TLD=' . self::$domain_without_tld );

			$production_db_to_staging_db = __DIR__ . '/../wp/sync_production_db_to_staging_db.sh';
			exec( "bash $production_db_to_staging_db", $dboutput, $dbwpout );
		}

		if ( $list_repos === self::$choice ) {
			self::$bitbucket_consumer_key    = self::$bitbucket_consumer_key ?: $io->askAndHideAnswer( '<info>Bitbucket consumer key</info>: ' );
			self::$bitbucket_consumer_secret = self::$bitbucket_consumer_secret ?: $io->askAndHideAnswer( '<info>Bitbucket consumer secret</info>: ' );
			self::$bitbucket_team_name       = self::$bitbucket_team_name ?: 'brewpixel';
			putenv( 'BITBUCKET_CONSUMER_KEY=' . self::$bitbucket_consumer_key );
			putenv( 'BITBUCKET_CONSUMER_SECRET=' . self::$bitbucket_consumer_secret );
			putenv( 'BITBUCKET_TEAM_NAME=' . self::$bitbucket_team_name );
			$list_bitbucket_repos = __DIR__ . '/list_bitbucket_repos.sh';
			exec( "bash $list_bitbucket_repos", $configoutput, $wp_config_result );
			print_r( $configoutput );
			var_dump( $wp_config_result );
		}
		if ( $install_reinstall_wordpress_current_project === self::$choice ) {
			self::questions( $io, self::$choice );

			putenv( 'CHOICE=' . self::$choice );
			putenv( 'BITBUCKET_TEAM_NAME=' . self::$bitbucket_team_name );
			putenv( 'BITBUCKET_PROJECT_NAME=' . self::$bitbucket_project_name );
			putenv( 'BITBUCKET_PROJECT_KEY=' . self::$bitbucket_project_key );
			putenv( 'BITBUCKET_PROJECT_DESCRIPTION=' . self::$bitbucket_project_description );
			putenv( 'CLIENT_NAME=' . self::$client_name );
			putenv( 'MAIN_REPO_NAME=' . self::$main_repo_name );
			putenv( 'THEME_NAME=' . self::$theme_name );
			putenv( 'THEME_NAME_LOWERCASE_HYPHEN_NO_SPACES=' . self::$theme_name_lowercase_hyphen_no_spaces );
			putenv( 'DB_NAME=' . self::$db_name );
			putenv( 'DB_USER=' . self::$db_user );
			putenv( 'DB_PASSWORD=' . self::$db_password );
			putenv( 'DB_HOST=' . self::$db_host );
			putenv( 'SITE_URL=' . self::$site_url );
			putenv( 'SITE_TITLE_BLOG_NAME=' . self::$site_title_blog_name );
			putenv( 'BLOG_DESCRIPTION=' . self::$blog_description );
			putenv( 'COMPOSER_PROJECT=' . self::$composer_project );
			putenv( 'IGNORE_THEMES=' . self::$ignore_themes );
			putenv( 'IGNORE_PLUGINS=' . self::$ignore_plugins );
			putenv( 'IGNORE_WP=' . self::$ignore_wp );
			putenv( 'ALL_FAKE_CONTENT=' . self::$all_fake_content );
			putenv( 'FAKE_COMMENTS=' . self::$fake_comments );
			putenv( 'FAKE_POSTS=' . self::$fake_posts );
			putenv( 'FAKE_CLIENT_USER=' . self::$fake_client_user );
			putenv( 'DEVELOPMENT_MODE=' . self::$development_mode );

			// @TODO We should check before taking actions that we have the
			// right info and the person wants us to proceed.
			$wp_create_config = __DIR__ . '/../wp/create_wordpress_config';
			exec( "bash $wp_create_config", $output, $wp_result );
			$wp_install_check = __DIR__ . '/../wp/install_wordpress_check.sh';
			exec( "bash $wp_install_check", $output, $wp_result );

			if ( 1 === $output ) {
				$wp = __DIR__ . '/../wp/install_wordpress';
				exec( "bash $wp", $output, $wpout );
			} elseif ( 1 === $wp_result ) {
				die( 'There was an error. Likely we can not connect to the database' );
			} else {
				$database_drop = $io->askConfirmation( '<info>The database already exists. Can we delete it?</info> ' );
				if ( true === $database_drop ) {
					$wp = __DIR__ . '/../wp/install_wordpress';
					exec( "bash $wp", $installoutput, $wpout );
				} else {
					die( 'You did not want to delete the database so we did not begin the WordPress installation.' );
				}
			}

			echo 'Finished installing WordPress';
			exec( "bash espeak 'Finished installing WordPress.' --stdout | aplay", $finoutput, $finout );
		}

		if ( $maintenance_update_core_plugins_themes_languages === self::$choice ) {
			if ( self::$maintenance ) {
				self::sshQuestions( $io, self::$choice );
				putenv( 'SSH_IP=' . self::$sship );
				putenv( 'SSH_USER=' . self::$sshuser );
				if ( file_exists( self::$project_root . '/maker/maker-config.php' ) ) {
					include self::$project_root . '/maker/maker-config.php';
					putenv( 'ADMIN_USER=' . self::$admin_user );
					// @TODO: If self::$main_repo_name is not set then we should ask for it.
					putenv( 'MAIN_REPO_NAME=' . self::$main_repo_name );
				}
				$extract                  = new Extract;
				// @TODO: If self::$main_repo_name is not set then we should ask for it.
				$domain_extract           = $extract->parse( self::$main_repo_name );
				self::$domain_without_tld = preg_replace( '/\.' . $domain_extract['suffix'] . '$/',
					'', $domain_extract );
				putenv( 'DOMAIN_WITHOUT_TLD=' . self::$domain_without_tld );

				$wpupdate = __DIR__ . '/../wp/update_wordpress';
				exec( "bash $wpupdate", $output, $wpout );
				/*
				 * Variables to incorporate into a script.
				 *
				 *  WordPress path: public_html/wp
					WordPress as a submodule: YES
					Other submodules: public_html/content/plugins/brewpress
					Locale

				@TODO: Save a report of updates.
				@TODO: Add no auto updates to wp-config.php
				@TODO: Add deploy methods git to SFTP etc.
				@TODO: Build script system to get current production database and bring into local and staging
				-
				Make sure you have a local copy of the website and it is set-up in your local development environment.

				Open up your local dev environment. Is it vagrant?

					vagrant up

				SSH into it if required such as for a vagrant box.

	                ssh vagrant@

				Check what git branch it is on

	                git branch -a

				Are there any files modified that should be committed or discarded?

					git status

				Yes, there are files that are modified and should be committed.

	                git add .

	                git commit -m 'Message'

				No, there aren't. Proceed to next step.

				Does the branch have commits different to master? Check if commmit hashes are the same. Check if git master is up to date with origin.

				On production:

	                git log -1

				Full example:

				Login to the production server and see if there are updates to be committed or downloaded.

				Maybe create a new "update-2016-05-01" branch from master?

				    git checkout -b maintenance/update-2016-06-02 origin/master
				Checklist:

				1. Ensure website files and database have recently been backed up
				    - Log in to backup server `ssh [backup_user]@[backup_server_ip]`
				    - `cd /home/[backup_user]/backup/archive/daily`
				    - Then cd into directory with the same day of the week as today to check that the backup is there.
				    - To make a manual backup of files `tar -czf /home/goldie/backups/initial_backup.tar.gz /srv/www/sonobex.com/www`
				    - @todo Check database
				    - To make a manual backup of all databases run `for I in $(mysql -uroot -e 'show databases' -s --skip-column-names); do mysqldump -uroot $I | gzip > "/home/goldie/mysqldump/$I.`date +\%Y-\%m-\%d`.sql.gz"; done` as root user. Or individually ` mysqldump -uroot sonobexcom | gzip > "/home/goldie/mysqldump/sonobexcom.`date +\%Y-\%m-\%d`.sql.gz"`
				2. On local development system `git checkout develop` or `git checkout -b update-2016-05-01` for example
				3. Update WordPress core:

				Script: (Update WordPress core on local files and not within Vagrant virtual machine unless you have permission rights set-up there.)

				@TODO: We may want to do a puppet run before hand or at least check that puppet has run that day.
				@TODO: Be aware of what user is running each command. Does it need sudo? Likely best to run most commands as the user for that website.
				@TODO: Example sudo -u pcginvestcoukstaging -i
				@TODO: For the end report we will want to know the list of plugins installed/active etc as well as what was updated. This way we can look at caching plugins or inactive etc.
				cd wp
			    git fetch
			    git fetch --tags
			    # If you want to check the latest tags
			    git tags
			    # Replace number with latest tag
			    git checkout 3.4
			    cd ../
			    git add wp
			    git commit -m 'Update to WordPress 3.4'
			    git push

				If "classic" WordPress set-up with wp-content, wp-admin, wp-includes in public directory:
					wp core update
				git add -A
				git commit -m 'Update WordPress'
					wp plugin update --all
				git add -A public_html/wp-content/plugins/wordpress-seo
				git add -A public_html/wp-content/plugins/other-plugins etc
					wp theme update --all
					wp core language update

				*/
			} else {
				echo 'This website does not have a BrewPixel maintenance package. To override add MAINTENANCE_PKG=true to .env';
			}
		}

		if ( $project_new_wordpress_genesis === self::$choice ||
		     $project_new_wordpress_genesis_woocommerce === self::$choice ||
		     $project_new_wordpress_moka === self::$choice ||
		     $project_new_wordpress_moka_woocommerce === self::$choice
		) {
			echo "Installing Genesis requirements\n";
			exec( 'cd ' . self::$project_root . ' && composer require \
			wpackagist-plugin/log-deprecated-notices:* \
			wpackagist-plugin/debug-bar:* \
			wpackagist-plugin/debug-bar-console:* \
			wpackagist-plugin/debug-bar-extender:* \
			wpackagist-plugin/debug-bar-shortcodes:* \
			wpackagist-plugin/debug-bar-constants:* \
			wpackagist-plugin/debug-bar-post-types:* \
			wpackagist-plugin/debug-bar-cron:* \
			wpackagist-plugin/debug-bar-actions-and-filters-addon:* \
			wpackagist-plugin/debug-bar-transients:* \
			wpackagist-plugin/debug-bar-list-dependencies:* \
			wpackagist-plugin/debug-bar-remote-requests:dev-trunk \
			wpackagist-plugin/developer:* \
			wpackagist-plugin/user-switching:* \
			wp-cli/wp-cli:* \
			--dev --no-update --no-interaction' );

			exec( 'cd ' . self::$project_root . ' && composer require \
			johnpbloch/wordpress:">=4.9.4" \
			brewpixel/brewpress:dev-master \
			wpackagist-plugin/gutenberg:dev-trunk \
			wpackagist-plugin/akismet:dev-trunk \
			--no-update --no-interaction' );

			putenv( 'CHILD_THEME=true' );
		}

		if ( $project_new_wordpress_genesis === self::$choice ||
		     $project_new_wordpress_genesis_woocommerce === self::$choice
		) {
			exec( 'cd ' . self::$project_root . ' && composer require \
			brewpixel/genesis:* \
			brewpixel/genesis-child-starter:dev-master \
			--no-update --no-interaction' );
		}

		if ( $project_new_wordpress_moka === self::$choice ||
		     $project_new_wordpress_moka_woocommerce === self::$choice
		) {
			exec( 'cd ' . self::$project_root . ' && composer require \
			brewpixel/moka:^0.1.0 \
			brewpixel/moka-child-starter:dev-master \
			--no-update --no-interaction' );
		}

		if ( $project_new_wordpress_genesis_woocommerce === self::$choice ||
		     $project_new_wordpress_moka_woocommerce === self::$choice
		) {
			exec( 'cd ' . self::$project_root . ' && composer require \
			wpackagist-plugin/woocommerce:dev-trunk \
			--no-update --no-interaction' );
		}

		if ( $project_new_wordpress_genesis === self::$choice ||
		     $project_new_wordpress_genesis_woocommerce === self::$choice ||
		     $project_new_wordpress_moka === self::$choice ||
		     $project_new_wordpress_moka_woocommerce === self::$choice ) {
			exec( 'cd ' . self::$project_root . " && composer update" );
			// language
			// choose plugins
			// import database
			self::questions( $io, self::$choice );

			// About to perform the following:
			// * Halt Vagrant (if running)
			// * Create directory selbycarpetsnew in /Users/goldie/Sites/vagrant-local/www
			// * Create files vvv-init.sh, wp-cli.yml, and vvv-hosts in directory selbycarpetsnew
			// * Create file selbycarpetsnew.conf in /Users/goldie/Sites/vagrant-local/config/nginx-config/sites
			// * Run `vagrant up --provision` to initialize site
			// Provisioning Vagrant will do the following:
			// * Create database selbycarpetsnew
			// * Install WordPress (release version) in the htdocs directory
			// * Make the site visible at selbycarpetsnew.local
			// Continue (y/n)? : y
			// @TODO Let's save a lot of this info like the consumer key in a database and then we can confirm rather than ask.
			// If the database is not available then we can fallback to asking.
			putenv( 'CHOICE=' . self::$choice );
			putenv( 'BITBUCKET_CONSUMER_KEY=' . self::$bitbucket_consumer_key );
			putenv( 'BITBUCKET_CONSUMER_SECRET=' . self::$bitbucket_consumer_secret );
			putenv( 'BITBUCKET_TEAM_NAME=' . self::$bitbucket_team_name );
			putenv( 'BITBUCKET_PROJECT_NAME=' . self::$bitbucket_project_name );
			putenv( 'BITBUCKET_PROJECT_KEY=' . self::$bitbucket_project_key );
			putenv( 'BITBUCKET_PROJECT_DESCRIPTION=' . self::$bitbucket_project_description );
			putenv( 'CLIENT_NAME=' . self::$client_name );
			putenv( 'MAIN_REPO_NAME=' . self::$main_repo_name );
			putenv( 'THEME_NAME=' . self::$theme_name );
			putenv( 'THEME_NAME_LOWERCASE_HYPHEN_NO_SPACES=' . self::$theme_name_lowercase_hyphen_no_spaces );
			putenv( 'DB_NAME=' . self::$db_name );
			putenv( 'DB_USER=' . self::$db_user );
			putenv( 'DB_PASSWORD=' . self::$db_password );
			putenv( 'DB_HOST=' . self::$db_host );
			putenv( 'SITE_URL=' . self::$site_url );
			putenv( 'SITE_TITLE_BLOG_NAME=' . self::$site_title_blog_name );
			putenv( 'BLOG_DESCRIPTION=' . self::$blog_description );
			putenv( 'COMPOSER_PROJECT=' . self::$composer_project );
			putenv( 'IGNORE_THEMES=' . self::$ignore_themes );
			putenv( 'IGNORE_PLUGINS=' . self::$ignore_plugins );
			putenv( 'IGNORE_WP=' . self::$ignore_wp );
			putenv( 'ALL_FAKE_CONTENT=' . self::$all_fake_content );
			putenv( 'FAKE_COMMENTS=' . self::$fake_comments );
			putenv( 'FAKE_POSTS=' . self::$fake_posts );
			putenv( 'FAKE_CLIENT_USER=' . self::$fake_client_user );
			putenv( 'DEVELOPMENT_MODE=' . self::$development_mode );

			exec( 'cd ' . self::$project_root . ' && composer update' );

			$install = __DIR__ . '/post_create_project_script.sh';
			echo 'about to install';

			exec( "bash $install", $output, $return_value );

			$wp_create_config = __DIR__ . '/../wp/create_wordpress_config';
			exec( "bash $wp_create_config", $output_config, $wp_result_config );
			$wp_install_check = __DIR__ . '/../wp/install_wordpress_check.sh';
			exec( "bash $wp_install_check", $output, $wp_result );

			if ( 1 === $output ) {
				$wp = __DIR__ . '/../wp/install_wordpress';
				shell_exec( "bash $wp" );
			} elseif ( 1 === $wp_result ) {
				die( 'There was an error. Likely we can not connect to the database' );
			} else {
				$database_name = exec( "wp eval 'echo DB_NAME;'" );
				if ( $database_name ) {
					$database_drop = $io->askConfirmation( "<info>The $database_name database already exists. Can we delete it?</info> " );
				} else {
					$database_drop = $io->askConfirmation( '<info>The database name is either undefined or does not yet exist. Continue to install and it will be created?</info> ' );
				}

				if ( true === $database_drop ) {
					$wp = __DIR__ . '/../wp/install_wordpress';
					exec( "bash $wp", $installoutput, $wpout );
				} else {
					die( 'You did not want to delete the database so we did not begin the WordPress installation.' );
				}
			}
			if ( $project_new_wordpress_genesis === self::$choice ||
			     $project_new_wordpress_genesis_woocommerce === self::$choice
			) {
				exec( 'cd ' . self::$project_root . ' && composer remove \
					brewpixel/genesis-child-starter:dev-master \
					--no-update --no-interaction' );
			}
		}
		// @TODO Would be nice to output some info as a summary of the project.
		// Maybe domain name, local URL, repos, database info etc.
		self::confirmationMessage();
		echo 'finished';
	}

	public static function sshQuestions( $io, $choice = null ) {
		if ( file_exists( self::$project_root . '/maker/maker-config.php' ) ) {
			include self::$project_root . '/maker/maker-config.php';
			putenv( 'ADMIN_USER=' . self::$admin_user );
			putenv( 'MAIN_REPO_NAME=' . self::$main_repo_name );
		}

		$choice = $choice ?: self::$choice;

		self::$main_repo_name                        = self::$main_repo_name ?: $io->askAndValidate(
			'<info>Domain name of website</info>: ',
			array( __CLASS__, 'validateURLwithoutProtocol' ),
			null,
			null
		);

		self::$sship                = self::$sship ?: $io->askAndValidate(
			'<info>SSH Server IP</info>: ',
			array( __CLASS__, 'validateIP' ),
			null,
			'176.58.116.55'
		);
		self::$sshuser               = self::$sshuser ?: $io->askAndValidate(
			'<info>SSH User</info>: ',
			array( __CLASS__, 'validate_debian_username' ),
			null,
			'goldie'
		);
	}

	public static function questions( $io, $choice = null ) {
		if ( file_exists( self::$project_root . '/maker/maker-config.php' ) ) {
			include self::$project_root . '/maker/maker-config.php';
			putenv( 'ADMIN_USER=' . self::$admin_user );
		}

		$choice       = $choice ?: self::$choice;
		self::$simple = $io->askConfirmation( '<info>Simple default install?</info> [<comment>y,n</comment>]? ' );
		if ( true === self::$simple ) {
			echo "Ok we'll ask as few questions as possible.\n";
		} else {
			echo "Ok, you're in control.";
		}
		self::$client_name                           = self::$client_name ?: $io->askAndValidate(
			'<info>Client company name</info>: ',
			array( __CLASS__, 'validate_client_name' ),
			null,
			null
		);
		self::$main_repo_name                        = self::$main_repo_name ?: $io->askAndValidate(
			'<info>Domain name of website</info>: ',
			array( __CLASS__, 'validateURLwithoutProtocol' ),
			null,
			null
		);
		self::$theme_name                            = self::$theme_name ?: $io->askAndValidate(
			'<info>New WordPress theme name</info>: ',
			array( __CLASS__, 'validate_theme_name' ),
			null,
			null
		);
		self::$theme_name_lowercase_hyphen_no_spaces = self::validate_lowercase_hyphen_no_spaces( self::$theme_name );
		if ( self::$simple ) {
			self::$db_name                       = self::$db_name ?: self::default_db_name( self::$main_repo_name . 'local' );
			self::$db_user                       = self::$db_user ?: 'root';
			self::$db_password                   = self::$db_password ?: 'root';
			self::$db_host                       = self::$db_host ?: 'localhost';
			self::$bitbucket_team_name           = self::$bitbucket_team_name ?: 'brewpixel';
			self::$bitbucket_project_name        = self::$bitbucket_project_name ?: self::$main_repo_name;
			self::$bitbucket_project_key         = self::$bitbucket_project_key ?: strtoupper( substr( self::$bitbucket_project_name,
				0, 3 ) );
			self::$bitbucket_project_description = self::$bitbucket_project_description ?: self::$main_repo_name . ' project.';

			$extract            = new Extract;
			$domain_extract     = $extract->parse( self::$main_repo_name );
			$domain_without_tld = preg_replace( '/\.' . $domain_extract['suffix'] . '$/',
				'', $domain_extract );

			self::$site_url             = self::$site_url ?: 'http://' . $domain_without_tld . '.local';
			self::$site_title_blog_name = self::$site_title_blog_name ?: 'My site';
			self::$blog_description     = self::$blog_description ?: 'A WordPress site by BrewPixel';
			self::$composer_project     = self::$composer_project ?: 1;
			self::$ignore_themes        = self::$ignore_themes ?: 1;
			self::$ignore_plugins       = self::$ignore_plugins ?: 1;
			self::$ignore_wp            = self::$ignore_wp ?: 1;
			self::$all_fake_content     = self::$all_fake_content ?: 1;
			self::$fake_comments        = self::$fake_comments ?: 1;
			self::$fake_posts           = self::$fake_posts ?: 1;
			self::$fake_client_user     = self::$fake_client_user ?: 1;
			self::$development_mode     = self::$development_mode ?: 1;
		} else {
			self::$db_name                = self::$db_name ?: $io->askAndValidate(
				'<info>Database name</info>: ',
				array( __CLASS__, 'validate_db_name' ),
				null,
				self::default_db_name( self::$main_repo_name . 'local' )
			);
			self::$db_user                = self::$db_user ?: $io->askAndValidate(
				'<info>Database user</info>: ',
				array( __CLASS__, 'validate_db_username' ),
				null,
				'root'
			);
			self::$db_password            = self::$db_password ?: $io->askAndHideAnswer( '<info>Database password</info>: ' );
			self::$db_host                = self::$db_host ?: $io->askAndValidate(
				'<info>Database host</info> (<comment>Default: localhost</comment>): ',
				array( __CLASS__, 'validate_lowercase_no_spaces' ),
				null,
				'localhost'
			);
			self::$bitbucket_team_name    = self::$bitbucket_team_name ?: $io->askAndValidate(
				'<info>Bitbucket team name</info> (<comment>Default: brewpixel</comment>): ',
				array( __CLASS__, 'validate_lowercase_no_spaces' ),
				null,
				'brewpixel'
			);
			self::$bitbucket_project_name = self::$bitbucket_project_name ?: $io->ask( '<info>Bitbucket project name</info>: ' );

			self::$bitbucket_project_key         = self::$bitbucket_project_key         ?: $io->ask( '<info>Bitbucket project key</info>: ' );
			self::$bitbucket_project_description = self::$bitbucket_project_description ?: $io->ask( '<info>Bitbucket project description</info>: ' );
			self::$site_url                      = self::$site_url                      ?: $io->ask( '<info>Local development site URL</info>: ' );
			self::$site_title_blog_name          = self::$site_title_blog_name          ?: $io->ask( '<info>Site title</info> (<comment>You can leave it blank for the default</comment>): ' );
			self::$blog_description              = self::$blog_description              ?: $io->ask( '<info>Blog description</info> (<comment>You can leave it blank for the default</comment>): ' );
			self::$composer_project              = self::$composer_project              ?: $io->askConfirmation( '<info>Is this a composer project?</info> [<comment>y,n</comment>]? ' );
			if ( true === self::$composer_project ) {
				self::$ignore_themes  = self::$ignore_themes  ?: $io->askConfirmation( "<info>Since this is a composer project we recommend to git ignore the WordPress themes. Is that ok?</info> [<comment>y,n</comment>]? \n(<comment>You can leave it blank for the default</comment>) " );
				self::$ignore_plugins = self::$ignore_plugins ?: $io->askConfirmation( "<info>Since this is a composer project we recommend to git ignore the WordPress plugins. Is that ok?</info> [<comment>y,n</comment>]? \n(<comment>You can leave it blank for the default</comment>) " );
				self::$ignore_wp      = self::$ignore_wp      ?: $io->askConfirmation( "<info>Since this is a composer project we recommend to git ignore the WordPress core. Is that ok?</info> [<comment>y,n</comment>]? \n(<comment>You can leave it blank for the default</comment>) " );
			} else {
				self::$ignore_themes  = self::$ignore_themes  ?: $io->askConfirmation( '<info>Git ignore WordPress themes?</info> [<comment>y,n</comment>]? (<comment>You can leave it blank for the default</comment>) ' );
				self::$ignore_plugins = self::$ignore_plugins ?: $io->askConfirmation( '<info>Git ignore WordPress plugins?</info> [<comment>y,n</comment>]? (<comment>You can leave it blank for the default</comment>) ' );
				self::$ignore_wp      = self::$ignore_wp      ?: $io->askConfirmation( '<info>Git ignore WordPress core?</info> [<comment>y,n</comment>]? (<comment>You can leave it blank for the default</comment>) ' );
			}

			self::$all_fake_content = self::$all_fake_content ?: $io->askConfirmation( '<info>All fake content?</info> [<comment>y,n</comment>]? ' );

			if ( true === self::$all_fake_content ) {
				self::$fake_comments    = 1;
				self::$fake_posts       = 1;
				self::$fake_client_user = 1;
			} else {
				self::$fake_comments    = self::$fake_comments    ?: $io->askConfirmation( '<info>Fake comments? </info> [<comment>y,n</comment>]? ' );
				self::$fake_posts       = self::$fake_posts       ?: $io->askConfirmation( '<info>Fake posts? </info> [<comment>y,n</comment>]? ' );
				self::$fake_client_user = self::$fake_client_user ?: $io->askConfirmation( '<info>Fake_client_user? </info> [<comment>y,n</comment>]? ' );
			}

			self::$development_mode = self::$development_mode ?: $io->askConfirmation( '<info>Development mode? </info> [<comment>y,n</comment>]? ' );

			// $result = preg_replace("/(^\\S+/{1}.+$)(?m)/uiUm", "hello $1", $searchText);
			// print_r($outputArray);
			// self::showPackages( $show );
			// Capture these into a list and ask if any are to be removed:
			// composer remove vendor/package vendor/package2
			$packages = [ 999 ];
			while ( 0 < $packages[0] && '' !== $packages[0] ) {

				exec( 'composer show --self', $array_packages,
					$return_value );

				$filtered_packages = preg_grep( '#(^\S+/{1}.+$)(?m)#',
					$array_packages );
				$root_packages     = array_merge( array( 'No' ),
					array_values( $filtered_packages ) );
				unset( $packages );
				$packages = $io->select(
					"<info>Select any packages you would like to remove:</info>\n<comment>(Multi-select comma separated numbers. Example 1,2,3)</comment> ",
					$root_packages,
					'No',
					false,
					'Value "%s" is invalid',
					true
				);

				if ( '0' !== $packages[0] && 'No' !== $packages[0] && '0' !== $packages && 'No' !== $packages ) {
					if ( '0' === $packages || 'No' === $packages || 'n' === $packages || '0' === $packages[0] || 'No' === $packages[0] || 'n' === $packages[0] ) {
						echo 'Exiting';
					} else {
						foreach ( $packages as $package ) {
							$package_remove = preg_replace( '/(^.*)\\s.*$/uiUm',
								'$1',
								$root_packages[ $package ] );
							exec( 'cd ' . self::$project_root . " && composer remove $package_remove --no-update --no-interaction" );
						}
					}
				}
				unset( $array_packages, $root_packages, $filtered_packages );
			}

			// Ask what packages they would like to add to dev-only:
			// composer require vendor/package:2.* vendor/package2:dev-master --dev --no-update
			$package = '';
			while ( 'No' !== $package && 'n' !== $package && 0 !== $package ) {
				$package = $io->askAndValidate(
					"<info>Add a package to use for development:</info>\n<comment>(This will be added to require-dev. Example vendor/package:dev-master)</comment> ",
					array( __CLASS__, 'validatePackage' ),
					null,
					'No'
				);

				if ( '0' === $package || 'No' === $package || 'n' === $package ) {
					echo 'Exiting';
				} else {
					exec( 'cd ' . self::$project_root . ' && composer require $package --dev --no-update --no-interaction' );
				}
			}
			unset( $package );

			// Ask what packages they would like to add for production as well:
			// composer require vendor/package:2.* vendor/package2:dev-master --no-update
			$package = '';
			while ( 'No' !== $package && 'n' !== $package && 0 !== $package ) {
				$package = $io->askAndValidate(
					"<info>Add a package to use for all environments:</info>\n<comment>(This will be added to require. Example vendor/package:dev-master)</comment> ",
					array( __CLASS__, 'validatePackage' ),
					null,
					'No'
				);

				if ( '0' === $package || 'No' === $package || 'n' === $package ) {
					echo 'Exiting';
				} else {
					exec( 'cd ' . self::$project_root . " && composer require $package --no-update --no-interaction" );
				}
			}

			// Ask if they want to search for any:
			// composer search monolog
		}
		self::$bitbucket_consumer_key    = self::$bitbucket_consumer_key ?: $io->askAndHideAnswer( '<info>Bitbucket consumer key</info>: ' );
		self::$bitbucket_consumer_secret = self::$bitbucket_consumer_secret ?: $io->askAndHideAnswer( '<info>Bitbucket consumer secret</info>: ' );

		/* TESTING / DEBUG */
		echo self::$choice . "\n";
		echo self::$main_repo_name . "\n";
		echo self::$theme_name . "\n";
		echo self::$theme_name_lowercase_hyphen_no_spaces . "\n";
		echo self::$db_name . "\n";
		echo self::$db_user . "\n";
		echo self::$db_password . "\n";
		echo self::$db_host . "\n";
		echo self::$bitbucket_consumer_key . "\n";
		echo self::$bitbucket_consumer_secret . "\n";
		echo self::$bitbucket_team_name . "\n";
		echo self::$bitbucket_project_name . "\n";
		echo self::$bitbucket_project_key . "\n";
		echo self::$bitbucket_project_description . "\n";
		echo self::$site_url . "\n";
		echo self::$site_title_blog_name . "\n";
		echo self::$blog_description . "\n";
		echo self::$composer_project . "\n";
		echo self::$ignore_themes . "\n";
		echo self::$ignore_plugins . "\n";
		echo self::$ignore_wp . "\n";
		echo self::$all_fake_content . "\n";
		echo self::$fake_comments . "\n";
		echo self::$fake_posts . "\n";
		echo self::$fake_client_user . "\n";
		echo self::$development_mode . "\n";
	}

	public static function confirmationMessage( $args = null ) {
		// @TODO Might be nice to show what is going to happen before doing it
		/*
		 * Example:
		 * About to perform the following:
		 * * Halt Vagrant (if running)
		 * * Create directory fhl-codename in /Users/goldie/Sites/vagrant-local/www
		 * * Create files vvv-init.sh, wp-cli.yml, and vvv-hosts in directory fhl-codename
		 * * Create file fhl-codename.conf in /Users/goldie/Sites/vagrant-local/config/nginx-config/sites
		 * * Run `vagrant up --provision` to initialize site
		 *
		 * Provisioning Vagrant will do the following:
		 * * Create database fhl-codename
		 * * Install WordPress (release version) in the htdocs directory
		 * * Make the site visible at fhl.local
		 *  Continue (y/n)? :
		 */
		echo 'Choice: ' . self::$choice . "\n";
		echo 'Main repo name: ' . self::$main_repo_name . "\n";
		echo 'Theme Name: ' . self::$theme_name . "\n";
		echo 'Theme directory name: ' . self::$theme_name_lowercase_hyphen_no_spaces . "\n";
		echo 'Database name: ' . self::$db_name . "\n";
		echo 'Database user: ' . self::$db_user . "\n";
		echo 'Database password: ' . self::$db_password . "\n";
		echo 'Database host: ' . self::$db_host . "\n";
		echo '' . self::$bitbucket_consumer_key . "\n";
		echo '' . self::$bitbucket_consumer_secret . "\n";
		echo 'Bitbucket team name: ' . self::$bitbucket_team_name . "\n";
		echo 'Bitbucket project name: ' . self::$bitbucket_project_name . "\n";
		echo 'Bitbucket project key: ' . self::$bitbucket_project_key . "\n";
		echo 'Bitbucket project description: ' . self::$bitbucket_project_description . "\n";
		echo 'Site URL: ' . self::$site_url . "\n";
		echo 'Site title aka "Blog name": ' . self::$site_title_blog_name . "\n";
		echo 'Blog description: ' . self::$blog_description . "\n";
		echo 'Composer: ' . self::$composer_project . "\n";
		echo 'Ignore themes: ' . self::$ignore_themes . "\n";
		echo 'Ignore plugins: ' . self::$ignore_plugins . "\n";
		echo 'Ignore WordPress: ' . self::$ignore_wp . "\n";
		echo 'All fake content: ' . self::$all_fake_content . "\n";
		echo 'Fake comments: ' . self::$fake_comments . "\n";
		echo 'Fake posts: ' . self::$fake_posts . "\n";
		echo 'Fake client user: ' . self::$fake_client_user . "\n";
		echo 'Development mode: ' . self::$development_mode . "\n";
	}

	public static function validatePackage( $args ) {
		if ( 'No' === $args || 'n' === $args ) {
			$args = 'No';
		}

		return $args;
	}

	public static function default_db_name( $string ) {

		$string = self::validate_db_name( $string );

		return $string;
	}

	public static function validate_client_name( $string ) {

		$sanitise = new Sanitise;
		// Limit theme name to 32 characters
		$string = $sanitise->limit_string_length( $string, 64 );
		$string = $sanitise->remove_disallowed_characters( $string );

		return $string;
	}

	public static function validate_theme_name( $string ) {

		$sanitise = new Sanitise;
		// Limit theme name to 32 characters
		$string = $sanitise->limit_string_length( $string, 32 );
		$string = $sanitise->remove_disallowed_characters( $string );

		return $string;
	}

	/**
	 * @param $url
	 *
	 * @return mixed|string
	 */
	public static function validateURLwithoutProtocol( $url ) {
		if ( ! $url || strlen( $url ) < 4 ) {
			return false;
		}
		$sanitise = new Sanitise;
		$url      = $sanitise->to_lowercase( $url );
		$url      = $sanitise->remove_spaces( $url );
		$url      = $sanitise->remove_disallowed_url_protocols( $url );
		$validate = new Validate;
		$url      = $validate->validate_url_without_protocol( $url );

		return $url;
	}

	/**
	 * @param $string
	 *
	 * @return mixed
	 */
	public static function validate_lowercase_no_spaces( $string ) {
		$sanitise = new Sanitise;
		$string   = $sanitise->to_lowercase( $string );
		$string   = $sanitise->remove_spaces( $string );
		$string   = $sanitise->remove_disallowed_characters( $string );

		return $string;
	}

	/**
	 * @param $string
	 *
	 * @return mixed
	 */
	public static function validate_lowercase_hyphen_no_spaces( $string ) {
		$sanitise = new Sanitise;
		$string   = $sanitise->to_lowercase( $string );
		$string   = $sanitise->replace_spaces( $string, '-' );
		$string   = $sanitise->remove_disallowed_characters( $string );

		return $string;
	}

	public static function validate_debian_username( $string ) {

		$sanitise = new Sanitise;
		// 16 characters for the username
		$string = $sanitise->limit_string_length( $string, 32 );

		$string = $sanitise->remove_spaces( $string );
		$string = $sanitise->remove_disallowed_characters( $string, array( ':', '\n', '\t', '/' ) );
		//$string = $sanitise->to_lowercase( $string );
		$string = $string[0] === '-' ? false : $string;

		return $string;
	}


	public static function validate_db_username( $string ) {

		$sanitise = new Sanitise;
		// 16 characters for the username
		$string = $sanitise->limit_string_length( $string, 16 );

		$string = $sanitise->to_lowercase( $string );
		$string = $sanitise->remove_spaces( $string );
		$string = $sanitise->remove_disallowed_characters( $string );

		return $string;
	}

	public static function validate_db_name( $string ) {

		// Cannot contain / \ or . characters
		// Cannot contain characters that are not permitted in file names.
		// Cannot end with space characters
		// Permitted characters in unquoted identifiers:
		// ASCII: [0-9,a-z,A-Z$_] (basic Latin letters, digits 0-9, dollar, underscore)
		// Extended: U+0080 .. U+FFFF
		// Permitted characters in quoted identifiers include the full Unicode Basic Multilingual Plane (BMP), except U+0000:
		// ASCII: U+0001 .. U+007F
		// Extended: U+0080 .. U+FFFF
		// ASCII NUL (U+0000) and supplementary characters (U+10000 and higher) are not permitted in quoted or unquoted identifiers.
		// Identifiers may begin with a digit but unless quoted may not consist solely of digits.
		// The identifier quote character is the backtick (“`”):
		// Database, table, and column names cannot end with space characters.
		// The identifier quote character is the backtick (“`”):
		$sanitise = new Sanitise;
		// Length: 64 (NDB storage engine: 63) So we will use 63 for safety.
		$string = $sanitise->limit_string_length( $string, 63 );

		$string = $sanitise->sanitise_with_preg_replace( $string, '/[^0-9,a-z,A-Z$_]/', '' );
		$string = $sanitise->to_lowercase( $string );
		$string = $sanitise->remove_spaces( $string );
		$string = $sanitise->remove_disallowed_characters( $string );

		return $string;
	}

	/**
	 * @param $url
	 *
	 * @return mixed|string
	 */
	public static function validateIP( $ip ) {

		//$sanitise = new Sanitise;
		//$url      = $sanitise->to_lowercase( $url );
		//$url      = $sanitise->remove_spaces( $url );
		//$url      = $sanitise->remove_disallowed_url_protocols( $url );
		$validate = new Validate;
		$validip      = $validate->validateIp( '', $ip );

		return $validip ? $ip : false;
	}

	/* Add db_name and user etc to wordpress config file before WordPress install */

	// public static function showPackages( ShowCommand $show = null ) {
	// new ShowCommand;
	// print_r( $show->getRootRequires() );
	// }
}
